"""
Database API
"""
from typing import Union
from abc import ABC, abstractmethod


BytesOrNone = Union[bytes, None]


class TreeMapStore(ABC):
    @abstractmethod
    def get_value(self, key: bytes) -> BytesOrNone:
        """
        Get a leaf value for the given key.
        Returns:
            - Value if the key exists, or
            - None if it's not found
        """
        pass

    @abstractmethod
    def get_node(self, key: bytes) -> BytesOrNone:
        """
        Get a node for the given key.
        Returns:
            - Node if the key exists, or
            - None if it's not found
        """
        pass

    @abstractmethod
    def set_value(self, key: bytes, value: bytes) -> bool:
        """
        Put a leaf key,value in the store. Overwrite the value for existing keys.
        Returns:
            - True if the write succeeds
            - False if it doesn't
        """
        pass

    @abstractmethod
    def set_node(self, key: bytes, value: bytes) -> bool:
        """
        Put a node value in the store. Overwrite the value for existing keys.
        Returns:
            - True if the write succeeds
            - False if it doesn't
        """
        pass

    @abstractmethod
    def delete_value(self, key) -> bool:
        """
        Delete a leaf value
         Returns:
            - True if the delete succeeds
            - False if it doesn't
        """
        pass

    @abstractmethod
    def delete_node(self, key) -> bool:
        """
        Delete a node
         Returns:
            - True if the delete succeeds
            - False if it doesn't
        """
        pass


class TreeMemoryStore(TreeMapStore):
    nodes: dict
    values: dict

    def __init__(self):
        self.nodes = {}
        self.values = {}

    def get_value(self, key: bytes) -> BytesOrNone:
        """
        Get a leaf value for the given key.
        Returns:
            - Value if the key exists, or
            - None if it's not found
        """
        return self.values.get(key, None)

    def get_node(self, key: bytes) -> BytesOrNone:
        """
        Get a node for the given key.
        Returns:
            - Node if the key exists, or
            - None if it's not found
        """
        return self.nodes.get(key, None)

    def set_value(self, key: bytes, value: bytes) -> bool:
        """
        Put a leaf key,value in the store. Overwrite the value for existing keys.
        Returns:
            - True if the write succeeds
            - False if it doesn't
        """
        self.values[key] = value
        return True

    def set_node(self, key: bytes, value: bytes) -> bool:
        """
        Put a node value in the store. Overwrite the value for existing keys.
        Returns:
            - True if the write succeeds
            - False if it doesn't
        """
        self.nodes[key] = value
        return True

    def delete_value(self, key) -> bool:
        """
        Delete a leaf value
         Returns:
            - True if the delete succeeds
            - False if it doesn't
        """
        if key not in self.values:
            return False
        del self.values[key]
        return True

    def delete_node(self, key) -> bool:
        """
        Delete a node
         Returns:
            - True if the delete succeeds
            - False if it doesn't
        """
        if key not in self.nodes:
            return False
        del self.nodes[key]
        return True



from smt.tree import SparseMerkleTree
from smt.utils import DEFAULTVALUE, PLACEHOLDER
from smt.proof import verify_proof


def test_tree_basics():
    tree = SparseMerkleTree()

    assert DEFAULTVALUE == tree.get(b"c")

    root1 = tree.update(b"a", b"a1")
    assert 32 == len(root1)
    assert root1 != PLACEHOLDER

    assert tree.update(b"b", b"b1")
    assert tree.update(b"c", b"c1")
    assert tree.update(b"d", b"d1")
    assert tree.update(b"e", b"e1")
    rootn = tree.update(b"f", b"f1")

    assert 32 == len(rootn)
    assert root1 != rootn

    assert b"a1" == tree.get(b"a")
    assert b"b1" == tree.get(b"b")
    assert b"c1" == tree.get(b"c")
    assert b"d1" == tree.get(b"d")
    assert b"e1" == tree.get(b"e")
    assert b"f1" == tree.get(b"f")
    assert b"" == tree.get(b"nope")

    rootn1 = tree.delete(b"c")
    assert 32 == len(rootn1)

    assert DEFAULTVALUE == tree.get(b"c")
    assert b"a1" == tree.get(b"a")
    assert b"b1" == tree.get(b"b")
    assert b"d1" == tree.get(b"d")
    assert b"e1" == tree.get(b"e")
    assert b"f1" == tree.get(b"f")

    # has it...
    assert tree.has(b"e")

    # update existing key
    rootn2 = tree.update(b"b", b"b11")
    assert 32 == len(rootn2)
    assert rootn2 != rootn1
    assert b"b11" == tree.get(b"b")

    # test updating a key still allows getting values from old roots
    rootn3 = tree.update(b"b", b"b111")
    assert b"b111" == tree.get(b"b")
    assert rootn3 != rootn2

    # test you can delete a key
    assert len(tree.delete(b"a")) > 0
    assert DEFAULTVALUE == tree.get(b"a")


def test_proofs():
    tree = SparseMerkleTree()
    assert tree.update(b"b", b"b1")
    assert tree.update(b"c", b"c1")
    assert tree.update(b"d", b"d1")
    assert tree.update(b"e", b"e1")

    assert tree.delete(b"c")

    root = tree.update(b"f", b"f1")

    proof = tree.prove(b"d")
    assert verify_proof(proof, root, b"d", b"d1")

    proof1 = tree.prove(b"np")
    assert not verify_proof(proof1, root, b"np", b"np1")

    proof2 = tree.prove(b"c")
    assert verify_proof(proof2, root, b"c", DEFAULTVALUE)


def test_bulk():
    data = make_random_data()
    tree = SparseMerkleTree()
    # Add data
    for k, v in data:
        assert tree.update(k, v)

    # Check it's there
    for k, v in data:
        assert v == tree.get(k)

    # Check proofs
    root = tree.root
    for k, v in data:
        proof = tree.prove(k)
        assert proof.sanity_check()
        assert verify_proof(proof, root, k, v)

    # Delete it
    for k, _ in data:
        assert tree.delete(k) != None

    # Check random value is deleted
    assert b"" == tree.get(data[4][0])


def make_random_data(size=500):
    import secrets
    import random

    return [
        (
            secrets.token_bytes(random.randint(10, 30)),
            secrets.token_bytes(random.randint(30, 500)),
        )
        for _ in range(size)
    ]



"""
Sparse Merkle Tree.
A Python port of: https://github.com/celestiaorg/smt
"""

from .proof import SparseMerkleProof
from .store import TreeMapStore, BytesOrNone, TreeMemoryStore
from .utils import (
    create_node,
    create_leaf,
    count_common_prefix,
    get_bit,
    parse_leaf,
    parse_node,
    is_leaf,
    digest,
    DEPTH,
    PLACEHOLDER,
    RIGHT,
    DEFAULTVALUE,
)

from typing import List, Sequence, Tuple


# Errors
KeyAlreadyEmpty = 1
InvalidKey = 2


class SparseMerkleTree:
    root: bytes
    store: TreeMapStore

    def __init__(
        self,
        store: TreeMapStore = TreeMemoryStore(),
        root=PLACEHOLDER,
    ):
        self.root = root
        self.store = store

    def root_as_bytes(self) -> bytes:
        return self.root

    def root_as_hex(self) -> str:
        return "0x{}".format(self.root.hex())

    def get(self, key: bytes) -> bytes:
        """
        Get a value for the given key using the current root.
        """
        if self.root == PLACEHOLDER:
            return DEFAULTVALUE

        path = digest(key)
        val = self.store.get_value(path)
        if not val:
            return DEFAULTVALUE

        return val

    def has(self, key):
        result = self.get(key)
        return result != DEFAULTVALUE

    def update(self, key, value):
        """
        Update a key value
        """
        new_root = self.update_for_root(key, value, self.root)
        self.root = new_root
        return new_root

    def update_for_root(self, key, value, root) -> BytesOrNone:
        path = digest(key)
        side_nodes, path_nodes, old_leafdata, _sibdata = self._get_sidenodes(
            path, root
        )

        # If the value is the None (defaultValue) then do a delete for the key
        if value == DEFAULTVALUE:
            new_root, err = self._delete_with_sidenodes(
                path, side_nodes, path_nodes, old_leafdata
            )
            if err and err == KeyAlreadyEmpty:
                return root
            if not self.store.delete_value(path):
                return None
            return new_root
        else:
            return self._update_with_sidenodes(
                path, value, side_nodes, path_nodes, old_leafdata
            )

    def delete(self, key):
        return self.update(key, DEFAULTVALUE)

    def delete_for_root(self, key, root):
        return self.update_for_root(key, DEFAULTVALUE, root)

    def prove(self, key):
        return self.prove_for_root(key, self.root)

    def prove_for_root(self, key, root):
        return self._proof_for_root(key, root, False)

    def prove_updatable(self, key):
        return self.prove_updatable(key, self.root)

    def prove_updatable_for_root(self, key, root):
        return self._proof_for_root(key, root, True)

    def _get_sidenodes(
        self, path: bytes, root: bytes, with_sibling_data=False
    ) -> Tuple[List[bytes], List[bytes], bytes, bytes]:
        """
        Walk the tree down from the root gathering neighbor nodes
        on the way to the leaf for the given key (path)
            parameters:
            returns:
        """
        side_nodes = []
        path_nodes = []
        path_nodes.append(root)

        if root == PLACEHOLDER:
            return (side_nodes, path_nodes, None, None)

        current_data = self.store.get_node(root)
        if current_data == None:
            return (None, None, None, None)
        elif is_leaf(current_data):
            return (side_nodes, path_nodes, current_data, None)

        node_hash = None
        side_node = None
        sibdata = None
        for i in range(DEPTH):
            l, r = parse_node(current_data)
            if get_bit(i, path) == RIGHT:
                side_node = l
                node_hash = r
            else:
                side_node = r
                node_hash = l

            side_nodes.append(side_node)
            path_nodes.append(node_hash)

            if node_hash == PLACEHOLDER:
                current_data = None
                break

            current_data = self.store.get_node(node_hash)
            if current_data == None:
                return (None, None, None, None)
            elif is_leaf(current_data):
                break

        if with_sibling_data:
            sibdata = self.store.get_node(side_node)
            if not sibdata:
                return (None, None, None, None)

        return (side_nodes[::-1], path_nodes[::-1], current_data, sibdata)

    def _update_with_sidenodes(
        self,
        path: bytes,
        value: bytes,
        side_nodes: Sequence[bytes],
        path_nodes: Sequence[bytes],
        old_leafdata: bytes,
    ) -> bytes:
        value_hash = digest(value)
        current_hash, current_data = create_leaf(path, value_hash)

        self.store.set_node(current_hash, current_data)  # = current_data
        current_data = current_hash

        common_prefix_count = 0
        old_value_hash = None
        if path_nodes[0] == PLACEHOLDER:
            common_prefix_count = DEPTH
        else:
            actual_path, old_value_hash = parse_leaf(old_leafdata)
            common_prefix_count = count_common_prefix(path, actual_path)

        if common_prefix_count != DEPTH:
            if get_bit(common_prefix_count, path) == RIGHT:
                current_hash, current_data = create_node(
                    path_nodes[0], current_data
                )
            else:
                current_hash, current_data = create_node(
                    current_data, path_nodes[0]
                )

            self.store.set_node(current_hash, current_data)  # = current_data
            current_data = current_hash
        elif old_value_hash != None:
            if old_value_hash == value_hash:
                return self.root

            self.store.delete_node(path_nodes[0])
            self.store.delete_value(path)

        for val in path_nodes[1:]:
            self.store.delete_node(val)

        offset = DEPTH - len(side_nodes)
        for i in range(DEPTH):
            side_node = bytes(32)
            if (i - offset) < 0:
                if (
                    common_prefix_count != DEPTH
                    and common_prefix_count > DEPTH - 1 - i
                ):
                    side_node = PLACEHOLDER
                else:
                    continue
            else:
                side_node = side_nodes[i - offset]

            if get_bit(DEPTH - 1 - i, path) == RIGHT:
                current_hash, current_data = create_node(
                    side_node, current_data
                )
            else:
                current_hash, current_data = create_node(
                    current_data, side_node
                )

            self.store.set_node(current_hash, current_data)  # = current_data
            current_data = current_hash

        self.store.set_value(path, value)
        return current_hash

    def _delete_with_sidenodes(self, path, sidenodes, path_nodes, old_leafdata):
        if path_nodes[0] == PLACEHOLDER:
            # This key is already empty as it is a placeholder; return an error
            return (None, KeyAlreadyEmpty)

        if parse_leaf(old_leafdata)[0] != path:
            # This key is already empty as a different key was found its place; return an error.
            return (None, KeyAlreadyEmpty)

        # Remove all orphans
        for val in path_nodes:
            if not self.store.delete_node(val):
                return (None, None)

        current_hash = None
        current_data = None
        non_placeholder_reached = False
        for i, sn in enumerate(sidenodes):
            if sn == None:
                continue

            if current_data == None:
                side_node_value = self.store.get_node(sn)
                if side_node_value == None:
                    return (None, InvalidKey)
                if is_leaf(side_node_value):
                    current_hash = sn
                    current_data = sn
                    continue
                else:
                    current_data = PLACEHOLDER
                    non_placeholder_reached = True

            if not non_placeholder_reached and sn == PLACEHOLDER:
                continue
            elif not non_placeholder_reached:
                non_placeholder_reached = True

            if get_bit(len(sidenodes) - 1 - i, path) == RIGHT:
                current_hash, current_data = create_node(sn, current_data)
            else:
                current_hash, current_data = create_node(current_data, sn)

            self.store.set_node(current_hash, current_data)  # = current_data
            current_data = current_hash

        if current_hash == None:
            current_hash = PLACEHOLDER

        return (current_hash, None)

    def _proof_for_root(self, key, root, is_updatable):
        path = digest(key)
        side_nodes, old_leafhash, old_leafdata, sibdata = self._get_sidenodes(
            path, root, is_updatable
        )

        non_empty_sides = []
        for sn in side_nodes:
            if sn != None:
                non_empty_sides.append(sn)

        non_membership_leafdata = None
        if old_leafhash != PLACEHOLDER:
            actual_path, _ = parse_leaf(old_leafdata)
            if path != actual_path:
                non_membership_leafdata = old_leafdata

        return SparseMerkleProof(
            non_empty_sides, non_membership_leafdata, sibdata
        )



"""
"""
from .utils import (
    DEPTH,
    KEYSIZE,
    create_leaf,
    digest,
    LEAF,
    parse_leaf,
    get_bit,
    create_node,
    DEFAULTVALUE,
    PLACEHOLDER,
    RIGHT,
)


class SparseMerkleProof:
    def __init__(self, sidenodes, non_membership_leafdata, siblingdata):
        self.sidenodes = sidenodes
        self.non_membership_leafdata = non_membership_leafdata
        self.sibling_data = siblingdata

    def sanity_check(self):
        if (
            len(self.sidenodes) > DEPTH
            or self.non_membership_leafdata != None
            and len(self.non_membership_leafdata)
            != len(LEAF) + KEYSIZE + KEYSIZE
        ):
            return False

        for sn in self.sidenodes:
            if len(sn) != KEYSIZE:
                return False

        if self.sibling_data:
            sibhash = digest(self.sibling_data)
            if self.sidenodes and len(self.sidenodes) > 0:
                if self.sidenodes[0] != sibhash:
                    return False
        return True


def verify_proof(proof, root, key, value):
    path = digest(key)

    if not proof.sanity_check():
        return False

    current_hash = None

    if value == DEFAULTVALUE:
        if not proof.non_membership_leafdata:
            current_hash = PLACEHOLDER
        else:
            actual_path, value_hash = parse_leaf(proof.non_membership_leafdata)
            if actual_path == path:
                return False
            current_hash, _current_data = create_leaf(actual_path, value_hash)
    else:
        value_hash = digest(value)
        current_hash, _current_data = create_leaf(path, value_hash)

    for i, node in enumerate(proof.sidenodes):
        if get_bit(len(proof.sidenodes) - 1 - i, path) == RIGHT:
            current_hash, _current_data = create_node(node, current_hash)
        else:
            current_hash, _current_data = create_node(current_hash, node)

    return current_hash == root



from hashlib import sha256

LEAF = b"\x00"
NODE = b"\x01"
KEYSIZE = 32
DEPTH = KEYSIZE * 8

RIGHT = 1
PLACEHOLDER = bytes(32)
DEFAULTVALUE = b""


def create_leaf(path, leaf_data):
    value = b"".join([LEAF, path, leaf_data])
    hash = digest(value)
    return (hash, value)


def parse_leaf(data):
    return (data[1:33], data[33:])


def is_leaf(data):
    return data[0] == 0


def create_node(left, right):
    value = b"".join([NODE, left, right])
    hash = digest(value)
    return (hash, value)


def parse_node(data):
    return (data[1:33], data[33:])


def digest(data):
    # we have to create a new instance has hashlib doesn't
    # have a 'reset' for updates
    hasher = sha256()
    hasher.update(data)
    return hasher.digest()


def get_bit(index, data):
    if data[index >> 3] & 1 << (7 - index % 8) > 0:
        return 1
    else:
        return 0


def count_set_bits(data):
    count = 0
    for i in range(0, len(data) * 8):
        if get_bit(i, data) == 1:
            count += 1
    return count


def count_common_prefix(a, b):
    count = 0
    for i in range(0, len(a) * 8):
        if get_bit(i, a) == get_bit(i, b):
            count += 1
        else:
            return count
    return count


# lil helper
def show_bits(value):
    x = [bin(byte) for byte in value]
    return "{}".format(x)


def big_test(tree, size):
    import secrets

    for i in range(size):
        k = secrets.token_bytes(10)
        v = bytes("hello{}".format(i), "utf-8")
        tree.update(k, v)
